use super::{
    get_player_item_or_fail, user_error, UResult, UserError, UserVerb, UserVerbRef, VerbContext,
};
use crate::{
    models::item::{Item, ItemSpecialData},
    static_content::room::{room_map_by_code, Direction},
};
use async_trait::async_trait;
use chrono::{Duration, Utc};
use itertools::Itertools;

pub struct Verb;
#[async_trait]
impl UserVerb for Verb {
    async fn handle(
        self: &Self,
        ctx: &mut VerbContext,
        _verb: &str,
        remaining: &str,
    ) -> UResult<()> {
        let item_name = remaining.trim();
        let player_item = get_player_item_or_fail(ctx).await?;
        let (loc_type, loc_code) = player_item
            .location
            .split_once("/")
            .ok_or_else(|| UserError("Invalid location".to_owned()))?;
        if loc_type != "room" {
            user_error(
                "You must go to where you rented the place (e.g. reception) to vacate.".to_owned(),
            )?;
        }
        let room = room_map_by_code()
            .get(loc_code)
            .ok_or_else(|| UserError("Can't find your room".to_owned()))?;
        if room.rentable_dynzone.is_empty() {
            user_error("Go to where you rented the place (e.g. reception) to vacate.".to_owned())?;
        }
        match room
            .rentable_dynzone
            .iter()
            .find(|ri| ri.rent_what == item_name)
        {
            None => user_error(format!(
                "Vacate must be followed by the specific thing you want to vacate: {}",
                room.rentable_dynzone
                    .iter()
                    .map(|ri| ri.rent_what)
                    .join(", ")
            ))?,
            Some(_) => (),
        };

        match ctx
            .trans
            .find_exact_dyn_exit(
                &player_item.location,
                &Direction::IN {
                    item: player_item.display.clone(),
                },
            )
            .await?
            .as_ref()
            .and_then(|it| it.location.split_once("/"))
        {
            None => user_error("You aren't renting anything from here!".to_owned())?,
            Some((ref ex_zone_t, ref ex_zone_c)) => {
                if let Some(ex_zone) = ctx
                    .trans
                    .find_item_by_type_code(ex_zone_t, ex_zone_c)
                    .await?
                {
                    match ex_zone.special_data {
                        Some(ItemSpecialData::DynzoneData {
                            vacate_after: Some(_),
                            ..
                        }) => user_error("Your lease is already up for termination.".to_owned())?,
                        Some(ItemSpecialData::DynzoneData {
                            vacate_after: None,
                            zone_exit: ref ex,
                        }) => {
                            ctx.trans
                                .save_item_model(&Item {
                                    special_data: Some(ItemSpecialData::DynzoneData {
                                        zone_exit: ex.clone(),
                                        vacate_after: Some(Utc::now() + Duration::days(1)),
                                    }),
                                    ..(*ex_zone).clone()
                                })
                                .await?;
                            ctx.trans.queue_for_session(ctx.session, Some("The robot files away your notice of intention to vacate. \"You have 24 hours to get all your stuff out, then the landlord will send someone up to boot out anyone still in there, and we will sell anything left behind to cover our costs. If you change your mind before then, just rent again and we'll cancel out your notice and let you keep the same apartment - then you'll have to pay the setup fee again though.\"\n")).await?
                        }
                        _ => user_error("The premises seem to be broken anyway".to_owned())?,
                    }
                }
            }
        }

        Ok(())
    }
}
static VERB_INT: Verb = Verb;
pub static VERB: UserVerbRef = &VERB_INT as UserVerbRef;
